#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <calypsi\intrinsics65816.h>
#include "MemAccess.h"
#include "SBC.h"
#include "BFIO.h"

#define dateSp(x)			((x < 10) ? "0" : "")
void printDir(char *path);

void clrRTC(void);

int main(void){
	char buf[201];
	uint8_t rets = 0;
	termreset();
	
	setVIA();
	
	printf("Initializing USB Controller\n");
	driverInit();
	
	
	while (!getDriveStatus()){
		if (checkIntMessage()) printf("Flash drive attached!\n");
	}
	
	
	printDir("/");
	
	// rtc_time tm;
	
	// writeSS1(0);
	// SPI_transfer(MCP795_CLRRAM);
	// SPI_transfer(0);
	// writeSS1(1);
	
	// writeSS1(0);
	// SPI_transfer(MCP795_CLRWDT);
	// SPI_transfer(0);
	// writeSS1(1);
	
	// if (!startRTC(0)){
		// printf("Couldn't start Oscillator!\n");
		// return -1;
	// }
	
	// setBitRTC(MCP795_REG_HOUR, MCP795_24_BIT, false);
	// setBitRTC(MCP795_REG_DAY, MCP795_VBATEN_BIT, true);
	
	// printf("\nReading Time!\n\n");
	// readTime(&tm);
	
	// printf("    %u-%s%u-%s%u %s%u:%s%u:%s%u\n", tm.tm_year+1900, dateSp(tm.tm_mon), tm.tm_mon, dateSp(tm.tm_mday), tm.tm_mday, dateSp(tm.tm_hour), tm.tm_hour, dateSp(tm.tm_min), tm.tm_min, dateSp(tm.tm_sec), tm.tm_sec);
	
	
	termreset();
	return 0;
}


// Format:
// 
// Directory: <Full Path>
// 
// <name>.<ext>    <size>    <last modified date/time>
// 
// Date/Time Format:
// YYYY-MM-DD HH:MM
// 
// Size format:
// 0000 *B

char *__prefix[] = {
	" B",
	"kB",
	"MB",
	"GB"
};

char *__spaces[] = {
	"",
	" ",
	"  ",
	"   ",
};

void printDir(char *path){
	char name[9];
	char ext[4];
	uint32_t fileS = 0, totalS = 0;
	uint16_t year;
	uint8_t month, day, hour, minute;
	uint8_t fileSPref = 0, leadSpaces = 0;
	uint8_t totalSPref = 0;
	
	name[8] = '\0';
	ext[3] = '\0';
	
	dirOpen(path);
	printf("Directory: \"%s\"\n\n", path);
	while(dirRead() == F_SUCCESS){
		
		// Split the filename into a seperate name and extension
		for (uint8_t i = 0; i < 8; i++){
			name[i] = __COD.fname[i];
		}
		ext[0] = __COD.fname[8];
		ext[1] = __COD.fname[9];
		ext[2] = __COD.fname[10];
		
		if (__COD.fattr & A_DIRECTORY){
			printf("%s    <DIR>           ", name);
		}else{
			
			// Get the File size
			fileS = __COD.fsize;
			totalS += fileS;
			
			// Adjust the File size to use kB/MB/GB if necessary and round down to the nearest fractional unit
			if (fileS > 1023U){				// kB Check
				fileS = fileS >> 10U;
				fileSPref = 1;
			}
			if (fileS > 1023U){				// MB Check
				fileS = fileS >> 10U;
				fileSPref = 2;
			}
			if (fileS > 1023U){				// GB Check
				fileS = fileS >> 10U;
				fileSPref = 3;
			}
			
			// Print the Name and Extension
			printf("%s     %s     ", name, ext);
			
			// Print leading spaces if necessary
			if (fileS < 1000) leadSpaces++;
			if (fileS < 100) leadSpaces++;
			if (fileS < 10) leadSpaces++;
			
			printf("%s%u %s", __spaces[leadSpaces], (uint16_t)fileS, __prefix[fileSPref]);
			
		}
		
		// Split the Last Modified Date to seperate variables
		year = dirGetYear(0);
		month = dirGetMonth(0);
		day = dirGetDay(0);
		hour = dirGetHour(0);
		minute = dirGetMinute(0);
		
		// Then Print it
		// Format: YYYY-MM-DD HH:MM
		printf("    %u-%s%u-%s%u %s%u:%s%u\n", year, dateSp(month), month, dateSp(day), day, dateSp(hour), hour, dateSp(minute), minute);
		
		
		fileS = 0;
		fileSPref = 0;
		leadSpaces = 0;
	}
	
	// Adjust the Size count to use kB/MB/GB if necessary and round down to the nearest fractional unit
	if (totalS > 1023U){			// kB Check
		totalS = totalS >> 10U;
		totalSPref = 1;
	}
	if (totalS > 1023U){			// MB Check
		totalS = totalS >> 10U;
		totalSPref = 2;
	}
	if (totalS > 1023U){			// GB Check
		totalS = totalS >> 10U;
		totalSPref = 3;
	}
	
	printf("\nTotal Size: %u %s\n", (uint16_t)totalS, __prefix[totalSPref]);
}



// -------------------------------------------------------------------------------- //
//                                                                                  //
// -------------------------------------------------------------------------------- //


file *fileOpen(char *path, uint8_t mode){
	
	// Try to open the File's path
	uint8_t tRet = openPath(path, true, true);
	
	// If the Mode is set to Create a File when it doesn't exist, and truncate
	// Don't bother to check if the file exists in the first place, since the command to create one
	// will overwrite any existing file anyways, saving some work
	if (mode & (O_CREATE | O_TRUNC)){
		writeSS0(0);
		SPI_transfer(FILE_CREATE);
		writeSS0(1);
	}
	
	// If the File cannot be opened, and the mode doesn't create a File, return early
	if ((!(mode & O_CREATE) && (tRet != F_SUCCESS)) || (tRet == F_DIR_NOT_FOUND)) return NULL;
	
	// If the Mode is set to Create a File when it doesn't exist, and NOT truncate
	// Check if the File already exists, and only create a new one if it doesn't
	if ((mode & O_CREATE) && !(mode & O_TRUNC) && (tRet == F_FILE_NOT_FOUND)){
		writeSS0(0);
		SPI_transfer(FILE_CREATE);
		writeSS0(1);
	}
	
	
	file *output = malloc(sizeof(file));
	output->fpath = malloc(strlen(path) * sizeof(char));
	strcpy(output->fpath, path);
	
	__COF = output;		// Set the current File as the Active one
	
	output->fmode = mode;
	output->ferror = F_SUCCESS;
	output->fptr = 0;
	
	if (path[0] == '/'){
		output->fpath = malloc(sizeof(char) * (strlen(path) + 1));
		strcpy(output->fpath, path);
	} else {
		output->fpath = malloc(sizeof(char) * (strlen(path) + strlen(__CWD) + 1));
		strcpy(output->fpath, __CWD);
		strcat(output->fpath, path);
	}
	
	writeSS0(0);
	SPI_transfer(GET_FILE_SIZE);
	SPI_transfer(0x68);
	output->fsize  = (uint32_t)SPI_read();
	output->fsize |= (uint32_t)SPI_read() << 8;
	output->fsize |= (uint32_t)SPI_read() << 16;
	output->fsize |= (uint32_t)SPI_read() << 24;
	writeSS0(1);
	
	if (mode & O_APPEND){
		output->fptr = output->fsize;
	}
	
	return output;
}


void fileClose(file *fp){
	if (checkIntMessage()){
		if (!getDriveStatus()){
			dbg_printf("DBG: fileClose: USB Drive Removed!\n");
			fp->ferror = F_ERROR;
			return;
		}
	}
	
	// Check if the File is currently open
	if (fp == __COF){
		// If Yes, close it
		writeSS0(0);
		SPI_transfer(FILE_CLOSE);
		SPI_transfer((fp->fmode & O_WRITE) ? 1 : 0);
		writeSS0(1);
		
		waitInterrupt();
		
		// Set the Active File to NULL
		__COF = NULL;
	}
	
	// Free the fpath and the FILE object
	free(fp->fpath);
	free(fp);
	
	fp->ferror = F_SUCCESS;
	return;
}


void fileSeek(file *fp, uint32_t value){
	fp->fptr = value;
	
	if (fp != __COF) return;
	
	writeSS0(0);
	SPI_transfer(BYTE_LOCATE);
	SPI_transfer(value);
	SPI_transfer((value >> 8));
	SPI_transfer((value >> 16));
	SPI_transfer((value >> 24));
	writeSS0(1);
}


void fileRead(file *fp, char *buf, uint16_t count){
	if (checkIntMessage()){
		if (!getDriveStatus()){
			dbg_printf("DBG: Read Error, USB Drive removed!\n");
			fp->ferror = F_ERROR;
			return;
		}
	}
	
	uint8_t tmp_ret;
	
	// Check if the File has to be opened first
	if (fp != __COF){
		dbg_printf("DBG: Changing the Active File!\n");
		
		// Close the Active File
		writeSS0(0);
		SPI_transfer(FILE_CLOSE);
		SPI_transfer((__COF->fmode & O_WRITE) ? 1 : 0);
		writeSS0(1);
		
		waitInterrupt();
		
		// Then Open the specified File
		tmp_ret = openPath(fp->fpath, true, true);
		if (tmp_ret != F_SUCCESS){
			dbg_printf("DBG: Something went horribly wrong!\n File at \"%s\" couldn't be opened!", fp->fpath);
			fp->ferror = tmp_ret;
			return;
		}
		
		// Move the File pointer to the correct location
		dbg_printf("DBG: Moving File Pointer to %u\n", (uint16_t)fp->fptr);
		writeSS0(0);
		SPI_transfer(BYTE_LOCATE);
		SPI_transfer(fp->fptr);
		SPI_transfer((fp->fptr >> 8));
		SPI_transfer((fp->fptr >> 16));
		SPI_transfer((fp->fptr >> 24));
		writeSS0(1);
	}
	
	
	bool reachedNL = false;
	uint8_t tiny_count, tmpByte;
	uint16_t buf_ptr = 0;
	
	// Request "count" amount of Bytes from the File
	writeSS0(0);
	SPI_transfer(BYTE_READ);
	SPI_transfer(count & 255U);
	SPI_transfer((count >> 8) & 255U);
	writeSS0(1);
	
	tmp_ret = waitInterrupt();					// Check what the CH376 has to say about this
	
	dbg_printf("DBG: Requesting %u Bytes from File: \"%s\"\n", count, fp->fpath);
	
	while (tmp_ret == USB_INT_DISK_READ){
		
		writeSS0(0);
		SPI_transfer(RD_USB_DATA0);				// Read a Block of data
		tiny_count = SPI_read();				// Amount of Bytes to read (always <256 Bytes)
		
		dbg_printf("DBG: Reading a Block! (%u Bytes)\n", tiny_count);
		
		for (uint16_t i = 0; i < tiny_count; i++){
			tmpByte = SPI_read();
			
			// Check if the Read Byte is a NewLine Character while reading in Text Mode
			if ((tmpByte == '\n') && !(fp->fmode & O_BINARY) && !reachedNL){
				dbg_printf("DBG: Read: NL/EOF Reached!\n");
				buf[buf_ptr + i] = '\0';		// Add a Null Terminator to the end of the buffer
				fp->fptr += buf_ptr + i + 1;	// Move the File Pointer 1 Character after the NewLine
				reachedNL = true;
			}
			// If so, continue to read the rest of the Block, but don't store any of it
			if (!reachedNL) buf[buf_ptr + i] = tmpByte;
		}
		writeSS0(1);
		
		if (reachedNL){
			writeSS0(0);
			SPI_transfer(BYTE_LOCATE);
			SPI_transfer(fp->fptr);
			SPI_transfer((fp->fptr >> 8));
			SPI_transfer((fp->fptr >> 16));
			SPI_transfer((fp->fptr >> 24));
			writeSS0(1);
			dbg_printf("DBG: Read: NL Reached!\n");
			fp->ferror = F_STOP_NL;
			return;
		}
		
		buf_ptr += tiny_count;					// Move the Pointer by how many bytes were read
		
		writeSS0(0);
		SPI_transfer(BYTE_RD_GO);				// Tell the CH376 to go to the next block
		writeSS0(1);
		
		tmp_ret = waitInterrupt();				// Check the interrupt message
	}
	
	fp->fptr += buf_ptr;						// Update the File Pointer
	
	if (buf_ptr < count){						// If the Pointer is smaller than the amount of Bytes Requested, EOF was reached early
		dbg_printf("DBG: Read: EOF Reached!\n");
		if (!(fp->fmode & O_BINARY)) buf[buf_ptr] = '\0';
		fp->ferror = F_STOP_EOF;
		return;
	} else{
		dbg_printf("DBG: Read: All Bytes Read!\n");
		if (!(fp->fmode & O_BINARY)) buf[buf_ptr] = '\0';
		fp->ferror = F_SUCCESS;
		return;
	}
}


void fileWrite(file *fp, char *buf, uint16_t count){
	
	// If Writing is not enabled for the File, return with an error
	if (!(fp->fmode & O_WRITE)){
		fp->ferror = F_NOT_WRITABLE;
		return;
	}
	
	if (checkIntMessage()){
		if (!getDriveStatus()){
			dbg_printf("DBG: Write Error, USB Drive removed!\n");
			fp->ferror = F_ERROR;
			return;
		}
	}
	
	uint8_t tmp_ret;
	
	// Check if the File has to be opened first
	if (fp != __COF){
		dbg_printf("DBG: Changing the Active File!\n");
		
		// Close the Active File
		writeSS0(0);
		SPI_transfer(FILE_CLOSE);
		SPI_transfer((__COF->fmode & O_WRITE) ? 1 : 0);
		writeSS0(1);
		
		waitInterrupt();
		
		// Then Open the specified File
		tmp_ret = openPath(fp->fpath, true, true);
		if (tmp_ret != F_SUCCESS){
			dbg_printf("DBG: Something went horribly wrong!\n File at \"%s\" couldn't be opened!", fp->fpath);
			fp->ferror = tmp_ret;
			return;
		}
		
		// Move the File pointer to the correct location
		dbg_printf("DBG: Moving File Pointer to %u\n", (uint16_t)fp->fptr);
		writeSS0(0);
		SPI_transfer(BYTE_LOCATE);
		SPI_transfer(fp->fptr);
		SPI_transfer((fp->fptr >> 8));
		SPI_transfer((fp->fptr >> 16));
		SPI_transfer((fp->fptr >> 24));
		writeSS0(1);
	}
	
	uint8_t tiny_count, tmpByte;
	uint16_t buf_ptr = 0;
	
	// Request writing "count" amount of Bytes to the File
	writeSS0(0);
	SPI_transfer(BYTE_WRITE);
	SPI_transfer(count & 255U);
	SPI_transfer((count >> 8) & 255U);
	writeSS0(1);
	
	tmp_ret = waitInterrupt();					// Check what the CH376 has to say about this
	
	dbg_printf("DBG: Requesting %u Bytes to File: \"%s\"\n", count, fp->fpath);
	
	while (tmp_ret == USB_INT_DISK_WRITE){
		
		writeSS0(0);
		SPI_transfer(WR_REQ_DATA);				// Write a Block of data
		tiny_count = SPI_read();				// Amount of Bytes to write (always <256 Bytes)
		
		dbg_printf("DBG: Writing a Block! (%u Bytes)\n", tiny_count);
		
		for (uint16_t i = 0; i < tiny_count; i++){
			SPI_transfer(buf[buf_ptr + i]);
		}
		writeSS0(1);
		
		buf_ptr += tiny_count;					// Move the Pointer by how many bytes were written
		
		writeSS0(0);
		SPI_transfer(BYTE_WR_GO);				// Tell the CH376 to go to the next block
		writeSS0(1);
		
		tmp_ret = waitInterrupt();				// Check the interrupt message
	}
	
	fp->fptr += buf_ptr;						// Update the File Pointer
	
	// If the File Pointer is beyond the size of the file, update the File Size
	if (fp->fsize < fp->fptr){
		dbg_printf("DBG: Updating the File Size! fptr (%u) is beyond fsize (%u)\n", fp->fptr, fp->fsize);
		fp->fsize = fp->fptr + 1;
		fileWrite(fp, NULL, 0);		// writing 0 Bytes tells the CH376 to write the new File Size to the USB Drive
	}
	
	fp->ferror = F_SUCCESS;
	return;
}


uint8_t fileDelete(char *path){
	// Open the file (and clear COF)
	uint8_t rets = openPath(path, true, true);
	__COF = NULL;
	
	// If successful, try to delete the specified file
	if (rets == F_SUCCESS){
		writeSS0(0);
		SPI_transfer(FILE_ERASE);
		writeSS0(1);
	}
	rets = waitInterrupt();
	
	if (rets == ANSW_USB_INT_SUCCESS) return F_SUCCESS;
	return F_ERROR;
}


uint8_t fileRename(char *path, char *newName){
	// oh god, the chip has no native solution. i have to:
	// 1. create a new file with the new name
	// 2. copy the contents of the old file into the new file
	// 3. delete the old file
	return F_ERROR;
}


uint8_t dirRead(){
	uint8_t tmp[32];
	char name[9];
	bool more = true;				// More files to read
	__COF = NULL;					// Clear the currently open file
	
	switch(__DR_STATE__){
		case ST_REQ:
			dbg_printf("DBG: Requesting a File from the Directory\n");
			
			// Then Open all files in the current Directory (using a wildcard)
			writeSS0(0);
			SPI_transfer(SET_FILE_NAME);
			SPI_transfer('*');
			SPI_transfer('\0');
			writeSS0(1);
			writeSS0(0);
			SPI_transfer(FILE_OPEN);
			writeSS0(1);
			
			__DR_ANSWER__ = waitInterrupt();
			
		case ST_READ:
			
			if (__DR_ANSWER__ == ANSW_ERR_MISS_FILE){
				dbg_printf("DBG: No more Files in the Directory\n");
				// No more files to read from the Directory
				more = false;
				
			}else if (__DR_ANSWER__ == ANSW_USB_INT_DISK_READ){
				dbg_printf("DBG: Reading from File\n");
				// Read 32 Bytes from the CH376 into "COD"
				writeSS0(0);
				SPI_transfer(RD_USB_DATA0);
				SPI_transfer(32U);
				for (uint8_t i = 0; i < 32U; i++){
					tmp[i] = SPI_read();
					if (i < 8) name[i] = tmp[i];
				}
				writeSS0(1);
				memcpy(&__COD, &tmp, 32U);
				
				name[8] = '\0';
				dbg_printf("DBG: File name: %s\n", name);
				
				// Then tell the CH376 to go to the next file
				writeSS0(0);
				SPI_transfer(FILE_ENUM_GO);
				writeSS0(1);
				
				__DR_ANSWER__ = waitInterrupt();
			}
		default:
		break;
	}
	
	if (more){
		__DR_STATE__ = ST_READ;
		return F_SUCCESS;
	}else{
		__DR_STATE__ = ST_REQ;
		return F_STOP_EOD;
	}
	
	
	// ST_REQ, 1. Open a File (FILE_OPEN, name: "*"), set state to ST_READ
	
	// ST_READ, 1. check if the file exists, if it doesn't set "more" to false, and skip to 3.
	// ST_READ, 2. if it does exist, read it's data (DIR_INFO_READ, 0xFF, RD_USB_DATA0) and go to the next file (FILE_ENUM_GO)
	// ST_READ, 3. set state to ST_DONE
	
	// 1. if "more" is false, set state to ST_REQ
	// 2. if it's true, set state to ST_READ
	// 3. return
}



// -------------------------------------------------------------------------------- //
//                                                                                  //
// -------------------------------------------------------------------------------- //


void delay10(uint16_t tms){
	uint32_t elapsed, now, start;
	
	start = timer();
	while(1){
		now = timer();
		elapsed = now - start;
		if (elapsed >= tms) return;
	}
}


void setVIA(void){
	mwrite8(_VIA_DDRA, 0b01111111);
	mwrite8(_VIA_ORA, _SPI_IDLE);
	mwrite8(_VIA_DDRB, 0b00000000);
}

void writeSS0(bool value){
	volatile uint8_t temp = mread8(_VIA_ORA);
	
	if (value){
		mwrite8(_VIA_ORA, _SPI_SS0 | temp);
	}else {
		mwrite8(_VIA_ORA, ~_SPI_SS0 & temp);
	}
}

void writeSS1(bool value){
	volatile uint8_t temp = mread8(_VIA_ORA);
	
	if (value){
		mwrite8(_VIA_ORA, _SPI_SS1 | temp);
	}else {
		mwrite8(_VIA_ORA, ~_SPI_SS1 & temp);
	}
}

uint8_t SPI_transfer(uint8_t value){
	volatile uint8_t out = 0, temp;
	
	for (int8_t i = 7; i > -1; i--){
		// Write the input bit to the MOSI line
		temp = mread8(_VIA_ORA);
		if (bitread(value, i)){
			mwrite8(_VIA_ORA, _SPI_MOSI | temp);
		}else {
			mwrite8(_VIA_ORA, ~_SPI_MOSI & temp);
		}
		
		// set the clock high
		temp = mread8(_VIA_ORA);
		mwrite8(_VIA_ORA, _SPI_SCK | temp);
		
		// read the MISO line and write the bit to the output
		bitwrite(out, i, (mread8(_VIA_ORA) & _SPI_MISO));
		
		// set the clock low
		temp = mread8(_VIA_ORA);
		mwrite8(_VIA_ORA, ~_SPI_SCK & temp);
	}
	dbg_printf("DBG: Sending: %X, Receiving: %X\n", value, out);
	return out;
}


uint8_t getInterrupt(void){
	writeSS0(0);
	SPI_transfer(GET_STATUS);
	uint8_t _tmpReturn = SPI_read();
	writeSS0(1);
	return _tmpReturn;
}

uint8_t waitInterrupt(void){
	__disable_interrupts();
	mwrite8(_CTRL, _CTRL_T1I);
	
	uint8_t time = 10;
	
	while(mread8(_VIA_ORB) & _SPI_INT){
		if (!time){
			return CH376_ERR_TIMEOUT;
		}
		__wait_for_interrupt();
		time--;
	}
	return getInterrupt();
}


bool pingDevice(void){
	bool output;
	writeSS0(0);
	SPI_transfer(CHECK_EXIST);
	SPI_transfer(0x55);
	output = (SPI_read() == 0xAA);
	writeSS0(1);
	return output;
}

bool initDevice(void){
	writeSS0(0);
	SPI_transfer(RESET_ALL);
	writeSS0(1);
	
	delay10(8);
	
	writeSS0(0);
	SPI_transfer(SET_USB_MODE);
	SPI_transfer(MODE_HOST_0);
	writeSS0(1);
	
	return pingDevice();
}

bool driverInit(void){
	while (!initDevice()) delay10(10);
	return true;
}

uint8_t mount(void){
	uint8_t tmpReturn = 0;
	writeSS0(0);
	SPI_transfer(DISK_MOUNT);
	writeSS0(1);
	
	tmpReturn = waitInterrupt();
	return tmpReturn;
}

void driveDetach(void){
	_ch376_deviceAttached = false;
}

void driveAttach(void){
	uint8_t tmpReturn = 0;
	
	writeSS0(0);
	SPI_transfer(SET_USB_MODE);
	SPI_transfer(MODE_HOST_1);
	writeSS0(1);
	writeSS0(0);
	SPI_transfer(SET_USB_MODE);
	SPI_transfer(MODE_HOST_2);		// Why set the mode twice?
	writeSS0(1);
	
	tmpReturn = waitInterrupt();
	
	if (tmpReturn == ANSW_USB_INT_CONNECT){
		for (uint8_t i = 0; i < 5; i++){
			tmpReturn = mount();
			
			if (tmpReturn == ANSW_USB_INT_SUCCESS){
				_ch376_deviceAttached = true;
				break;
			}else if (tmpReturn != CH376_ERR_TIMEOUT){
				break;
			}
		}
	} else driveDetach();
	
	__CWD[0] = '/';		// Set the CWD to the root directory
	__CWD[1] = 0x00;
}



bool checkIntMessage(void){
	uint8_t tmpReturn = 0;
	bool intRequest = false;
	
	while (!(mread8(_VIA_ORB) & _SPI_INT)){
		tmpReturn = getInterrupt();
		delay10(1);
	}
	switch (tmpReturn){
		case ANSW_USB_INT_CONNECT:
			intRequest = true;
			driveAttach();
			break;
		case ANSW_USB_INT_DISCONNECT:
			intRequest = true;
			driveDetach();
			break;
		}
	return intRequest;
}




// Open the path to a file, and optionally opens the final file as well
uint8_t openPath(char *path, bool open, bool chcwd){
	uint16_t str_ptr = 0;				// Index into the input string
	bool file = false;					// Specifies if the current try is on a directory or file
	char temp[14];						// Create a temporary string
	
	dbg_printf("DBG: Trying to Open File \"%s\"\n", path);
	
	// If the first character is a Slash, start from the Root directory
	if (path[0] == '/'){
		
		dbg_printf("DBG: Starting from the Root Directory\n");
		
		str_ptr++;
		writeSS0(0);
		SPI_transfer(SET_FILE_NAME);
		SPI_transfer('/');
		SPI_transfer('\0');
		writeSS0(1);
		writeSS0(0);
		SPI_transfer(FILE_OPEN);
		writeSS0(1);
		
		_ch376_answer = waitInterrupt();
		if (_ch376_answer != ANSW_ERR_OPEN_DIR){
			dbg_printf("DBG: What the Fuck? Couldn't Open Root Directory\n");
			return F_ERROR;
		}
		
		if (chcwd){
			__CWD[0] = '/';
			__CWD[1] = '\0';
		}
	}
	
	while (path[str_ptr]){
		for (uint8_t i = 0; i < 13; i++){	// Copy part of the input string into temp
			if (path[str_ptr] == '/'){		// If the character is a Slash, write a null byte to temp and exit
				temp[i] = '\0';
				str_ptr++;
				break;
			}
			temp[i] = path[str_ptr];
			if (!path[str_ptr]){			// If the end of the string has been reached, exit
				file = true;
				break;
			}
			str_ptr++;
		}
		dbg_printf("DBG: Opening: \"%s\"\n", temp);
		
		// Send the File/Directory Name to the CH376
		writeSS0(0);
		SPI_transfer(SET_FILE_NAME);
		for (uint8_t i = 0; i < 13; i++){
			SPI_transfer(temp[i]);
			if (!temp[i]) break;
		}
		writeSS0(1);
		
		// Try to open the specified File/Directory
		// But only if it's a directory, or if the "open" argument is >0 also files
		if (!file || open){
			writeSS0(0);
			SPI_transfer(FILE_OPEN);
			writeSS0(1);
			_ch376_answer = waitInterrupt();
		}else{
			// File reached, but "open" argument said not to open it. so return
			return F_SUCCESS;
		}
		
		// If the File has been opened, exit with success
		if (_ch376_answer == ANSW_USB_INT_SUCCESS){
			dbg_printf("DBG: File Opened!\n");
			return F_SUCCESS;
		}
		
		// If the File/Directory wasn't found, exit with an Error
		if (_ch376_answer == ANSW_ERR_MISS_FILE){
			if (file){
				dbg_printf("DBG: File not Found!\n");
				return F_FILE_NOT_FOUND;
			}else{
				dbg_printf("DBG: Directory not Found!\n");
				return F_DIR_NOT_FOUND;
			}
		}
		
		// If a Directory was opened, update the cwd and go for another round
		if (_ch376_answer == ANSW_ERR_OPEN_DIR){
			dbg_printf("DBG: Directory Opened!\n");
			if (chcwd){
				if ((strlen(__CWD) + strlen(temp)) > 255) return F_PATH_TOO_LONG;
				strcat(__CWD, temp);
			}
		}
	}
	return F_ERROR;
}




void readRTC(uint8_t addr, uint8_t *buf, uint8_t count){
	dbg_printf("DBG: readRTC: Reading %u bytes from address %X!\n", count, addr);
	
	// Send the Command to read from the RTC
	writeSS1(0);
	SPI_transfer(MCP795_READ);
	SPI_transfer(addr);
	
	// Then read as many bytes as requested
	for (uint8_t i = 0; i < count; i++){
		buf[i] = SPI_read();
	}
	writeSS1(1);
}

void writeRTC(uint8_t addr, uint8_t *buf, uint8_t count){
	dbg_printf("DBG: writeRTC: Writing %u bytes to address %X!\n", count, addr);
	
	// Send the Command to write to the RTC
	writeSS1(0);
	SPI_transfer(MCP795_WRITE);
	SPI_transfer(addr);
	
	// Then read as many bytes as requested
	for (uint8_t i = 0; i < count; i++){
		SPI_transfer(buf[i]);
	}
	writeSS1(1);
}

void setBitRTC(uint8_t addr, uint8_t bit, bool val){
	uint8_t tmp;
	
	// Read the Byte from the RTC
	readRTC(addr, &tmp, 1);
	dbg_printf("DBG: setBitRTC: read: %X, bit to %s: %u\n", tmp, val ? "set" : "clear", bit);
	
	// First check if the bit is already set to the requested value
	if (!bitread(tmp, bit) != !val){
		
		// If not, set/clear the specified bit and write it back to the RTC
		bitwrite(tmp, bit, val);
		dbg_printf("DBG: setBitRTC: write: %X\n", tmp);
		writeRTC(addr, &tmp, 1);
	}
}


bool stopRTC(bool *extosc){
	uint8_t data;
	uint8_t tries = 5;
	
	// Disable the RTC Oscillator
	setBitRTC(MCP795_REG_SECONDS, MCP795_ST_BIT, false);
	
	// Save a copy of the EXTOSC bit
	readRTC(MCP795_REG_CONTROL, &data, 1);
	*extosc = !!(data & MCP795_EXTOSC_BIT);
	
	// Clear the EXTOSC bit
	setBitRTC(MCP795_REG_CONTROL, MCP795_EXTOSC_BIT, false);
	
	// wait for the OSCON bit to be cleared
	do{
		readRTC(MCP795_REG_DAY, &data, 1);
		if (!(data & MCP795_OSCON_BIT)) break;
	}while(--tries);
	
	return tries;
}

bool startRTC(bool *extosc){
	uint8_t data;
	uint8_t tries = 5;
	
	// Check if the EXTOSC was previously set, and only set it if it was
	if (*extosc){
		setBitRTC(MCP795_REG_CONTROL, MCP795_EXTOSC_BIT, true);
	}
	
	// Enable the RTC Oscillator
	setBitRTC(MCP795_REG_SECONDS, MCP795_ST_BIT, true);
	
	// wait for the OSCON bit to be set
	do{
		readRTC(MCP795_REG_DAY, &data, 1);
		if (data & MCP795_OSCON_BIT) break;
	}while(--tries);
	
	return tries;
}





void readTime(rtc_time *tm){
	uint8_t data[7];
	
	readRTC(MCP795_REG_SECONDS, data, 7);
	
	tm->tm_sec =  bcd2bin(data[0] & 0x7F);
	tm->tm_min =  bcd2bin(data[1] & 0x7F);
	tm->tm_hour = bcd2bin(data[2] & 0x3F);
	tm->tm_wday = bcd2bin(data[3] & 0x07);
	tm->tm_mday = bcd2bin(data[4] & 0x3F);
	tm->tm_mon =  bcd2bin(data[5] & 0x1F);
	tm->tm_year = bcd2bin(data[6]) + 100;		// Assume year 20xx
}

void writeTime(rtc_time *tm){
	uint8_t data[7];
	bool extosc;
	
	// Stop the Oscillator
	stopRTC(&extosc);
	
	// Read first so that the config bits remain untouched
	readRTC(MCP795_REG_SECONDS, data, 7);
	
	// Then Prepare the Data
	data[0] = (data[0] & 0x80) | bin2bcd(tm->tm_sec);
	data[1] = (data[1] & 0x80) | bin2bcd(tm->tm_min);
	data[2] = bin2bcd(tm->tm_hour);
	data[3] = (data[3] & 0xF8) | bin2bcd(tm->tm_wday);
	data[4] = bin2bcd(tm->tm_mday);
	data[5] = (data[5] & MCP795_LP_BIT) | bin2bcd(tm->tm_mon);
	
	if (tm->tm_year > 100) tm->tm_year -= 100;
	
	data[6] = bin2bcd(tm->tm_year);
	
	// And write (most of) it back to the RTC
	writeRTC(MCP795_REG_SECONDS, data, 5);
	
	// The Month and Year are done seperately
	writeRTC(MCP795_REG_MONTH, &data[5], 2);
	
	// And finally start the Oscillator again
	startRTC(&extosc);
}

// 1;
// 50;
// 15;
// 7;
// 25;
// 6;
// 123;


